#include <cstddef>
using std::size_t;

#include <cstdlib>
using std::exit;

#include <cassert>

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <iostream>
using std::cerr;
using std::endl;

#include <utility>
using std::pair;
using std::make_pair;

#include <vector>
using std::vector;

const size_t L = 100;
const double d = 0.75/L;
int occupied[L][L];
const int delay = 50; // milliseconds

#include <random>
using std::mt19937;
using std::uniform_real_distribution;
using std::uniform_int_distribution;

#include <chrono>
using std::chrono::high_resolution_clock;

#include <functional>
using std::bind;

#include <algorithm>
using std::shuffle;

static mt19937 rng; // mersenne twistor random number generator
static auto rnd = bind(uniform_real_distribution<double>(0,1),rng); // random number in [0,1)
static auto rndi = bind(uniform_int_distribution<size_t>(0,L-1),rng); // random number in [0,L)

void display()
{
	glClear(GL_COLOR_BUFFER_BIT);
	for (size_t i = 0; i < L; ++i)
		for (size_t j = 0; j < L; ++j)
			if (occupied[i][j] > 0)
			{
				const double x = 2.0*(i+0.5)/L - 1.0;
				const double y = 2.0*(j+0.5)/L - 1.0;
                if (occupied[i][j] == 1)
                    glColor3f(1.0, 1.0, 1.0);
                else
                {
                    assert(occupied[i][j] == 2);
                    glColor3f(1.0, 0.5, 0.25);
                }
				glBegin(GL_POLYGON);
					glVertex2f(x-d,y-d);
					glVertex2f(x-d,y+d);
					glVertex2f(x+d,y+d);
					glVertex2f(x+d,y-d);
				glEnd();
				glFlush();
			}
	glutSwapBuffers();
}

void keyboard(unsigned char key, int x, int y)
{
	if (key == 'q')
		exit(0);
}

inline void if_stationary_mark(size_t i, size_t j)
{
    if (occupied[i][j] == 1 and
        (
         occupied[(i+1)%L][j] > 0 or occupied[(i+L-1)%L][j] > 0 or
         occupied[i][(j+1)%L] > 0 or occupied[i][(j+L-1)%L] > 0 or
         occupied[(i+1)%L][(j+1)%L] > 0 or occupied[(i+1)%L][(j+L-1)%L] > 0 or
         occupied[(i+L-1)%L][(j+1)%L] > 0 or occupied[(i+L-1)%L][(j+L-1)%L] > 0
        ))
        occupied[i][j] = 2; // 2:on but not a singleton (has neighbours)
}

double pop = -1.0, dep = -1.0;
void initialize_grid(void)
{
	for (size_t i = 0; i < L; ++i)
		for (size_t j = 0; j < L; ++j)
            occupied[i][j] = ( rnd () < pop ? 1 : 0 ); // 0:off and 1:on

     for (size_t i = 0; i < L; ++i)
        for (size_t j = 0; j < L; ++j)
            if_stationary_mark(i,j);
}


void sweep_grid(void)
{
    static vector< pair<size_t, size_t> > moveable;
    moveable.clear();
    
	for (size_t i = 0; i < L; ++i)
		for (size_t j = 0; j < L; ++j)
		{
            if (occupied[i][j] == 1)
                moveable.push_back(make_pair(i,j));
        }

    shuffle(moveable.begin(),moveable.end(),rng);
    
    for (auto coord : moveable)
    {
        const size_t i = coord.first;
        const size_t j = coord.second;
        if_stationary_mark(i,j);
        if (occupied[i][j] == 1) // is a moveable singleton
        {
            occupied[i][j] = 0;
            const double r = rnd();
            if (r < 0.25)
                occupied[(i+1)%L][j] = 1;
            else if (r < 0.5)
                occupied[(i+L-1)%L][j] = 1;
            else if (r < 0.75)
                occupied[i][(j+1)%L] = 1;
            else
                occupied[i][(j+L-1)%L] = 1;
        }
    }
    
    const int count = int(dep+1);
    for (int c = 0; c < count; ++c)
    {
        if (rnd() < dep/count)
        {
            const size_t i = rndi();
            const size_t j = rndi();
            if (occupied[i][j] == 0)
                occupied[i][j] = 1;
        }
    }
    
    for (size_t i = 0; i < L; ++i)
        for (size_t j = 0; j < L; ++j)
            if_stationary_mark(i,j);
}

void animate(int)
{
	sweep_grid();
	glutPostRedisplay();
	glutTimerFunc(delay,animate,0);
}

int main(int argc, char **argv)
{
    if (argc != 3)
    {
    hard_exit:
        cerr << "diffusion (0.0 < initial population probability < 1.0) (0.0 < deposition rate)" << endl;
        exit(1);
    }
    
    pop = atof(argv[1]);
    if (pop < 0.0 or pop > 1.0) goto hard_exit;
    
    dep = atof(argv[2]);
    if (dep < 0.0) goto hard_exit;
    
    rng.seed(mt19937::result_type(high_resolution_clock::now().time_since_epoch().count()));
    
	glutInit(&argc,argv);
	glutInitDisplayMode (GLUT_RGB | GLUT_DOUBLE);
	glutInitWindowSize (640, 640);
	glutInitWindowPosition (100, 100);
	glutCreateWindow("The DDA model");
	glutDisplayFunc(display);
    glutKeyboardFunc(keyboard);
	glutTimerFunc(delay,animate,0);
	
	initialize_grid();
	glutMainLoop();
}

