#ifndef POLYMER_HPP
#define POLYMER_HPP

#include <cstddef>
using std::size_t;

#include <cassert>

#include <deque>
using std::deque;

#include <algorithm>
using std::max;
using std::swap;

#include <limits>
using std::numeric_limits;

class monomer
{
public:
	long int x;
	long int y;
	monomer(size_t x_, size_t y_) : x(x_), y(y_) {}
};

class polymer
{
private:
	deque<monomer> chain;
public:
	polymer() : chain() {}
	void initialize_stretched(size_t L)
	{ 
		for (size_t i = 0; i <= L; ++i)
			chain.push_back(monomer(i,0));
	}
	template <typename Func>
	void initialize_wiggle(size_t L, Func &Rand)
	{ 
		long int x = 0;
		long int y = 0;
		for (size_t i = 0; i <= L; ++i)
		{
			if (Rand() < 0.5)
				++x;
			else
				++y;
			chain.push_back(monomer(x,y));
		}
	}
	typedef deque<monomer>::const_iterator const_iterator;
	typedef deque<monomer>::iterator iterator;
	iterator begin() { return chain.begin(); }
	iterator end() { return chain.end(); }
	size_t size() { return chain.size(); }
	size_t length() { assert(chain.size() > 1); return chain.size()-1; }
	long int distance2()
	{
		const long int dx = chain.front().x - chain.back().x;
		const long int dy = chain.front().y - chain.back().y;
		return dx*dx + dy*dy;
	}
	void query_dimensions(double &cmx, double &cmy, long int &wx, long int &wy)
	{
		long int x = 0;
		long int y = 0;
		long int max_x = -numeric_limits<long int>::max();
		long int min_x = numeric_limits<long int>::max();
		long int max_y = -numeric_limits<long int>::max();
		long int min_y = numeric_limits<long int>::max();
		for (iterator step = begin(); step < end(); ++step)
		{
			x += step->x;
			y += step->y;
			if (step->x > max_x) max_x = step->x;
			if (step->x < min_x) min_x = step->x;
			if (step->y > max_y) max_y = step->y;
			if (step->y < min_y) min_y = step->y;
		}
		assert(max_x >= min_x);
		assert(max_y >= min_y);
		cmx = double(x)/size();
		cmy = double(y)/size();
		wx = max_x - min_x;
		wy = max_y - min_y;
	}
	template <typename Func>
	void reptation_orw(Func &Rand) 
	{
		long int x = chain.back().x;
		long int y = chain.back().y;
		const double R = Rand();
		if (R < 0.25)
			x += 1;
		else if (R < 0.5)
			x -= 1;
		else if (R < 0.75)
			y += 1;
		else
			y -= 1;
		chain.push_back(monomer(x,y));
		chain.pop_front();		
	}
};

#endif // POLYMER_HPP
